package com.dfsek.feldspar.state

import android.util.Log
import androidx.compose.runtime.MutableState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.setValue
import androidx.lifecycle.LifecycleOwner
import arrow.core.Either
import com.dfsek.feldspar.util.TimelineEventWrapper
import org.matrix.android.sdk.api.session.Session
import org.matrix.android.sdk.api.session.events.model.*
import org.matrix.android.sdk.api.session.room.Room
import org.matrix.android.sdk.api.session.room.sender.SenderInfo
import org.matrix.android.sdk.api.session.room.timeline.*

class ChatRoomState(
    val room: Room,
    val session: Session
) : Timeline.Listener {
    private var timeline: Timeline? by mutableStateOf(null)
    var replyTo: TimelineEvent? by mutableStateOf(null)
    var selectedImageEvent: TimelineEvent? by mutableStateOf(null)
    private var timelineEvents: List<TimelineEvent> by mutableStateOf(emptyList())
    var splitEvents: MutableState<List<Pair<SenderInfo, Either<List<TimelineEventWrapper>, TimelineEventWrapper>>>> = mutableStateOf(emptyList())

    override fun onTimelineUpdated(snapshot: List<TimelineEvent>) {
        timelineEvents = snapshot.reversed()
        splitEvents.value = splitEvents()
    }


    fun startSync() {
        timeline = room.timelineService()
            .createTimeline(null, TimelineSettings(initialSize = 35)).also {
                it.addListener(this)
                it.start()
            }
    }

    fun stopSync() {
        timeline?.let {
            it.removeAllListeners()
            it.dispose()
        }
        timeline = null
    }

    private fun splitEvents(): List<Pair<SenderInfo, Either<List<TimelineEventWrapper>, TimelineEventWrapper>>> {
        val list = mutableListOf<Pair<SenderInfo, Either<MutableList<TimelineEventWrapper>, TimelineEventWrapper>>>()

        val redactionEventIDs = mutableSetOf<String>()
        val redactedEventIDs = mutableSetOf<String>()

        val redactedBy = mutableMapOf<String, TimelineEvent>()

        val replace = mutableMapOf<String, TimelineEvent>()

        timelineEvents.forEach {
            if (it.root.getClearType() == "m.room.redaction") {
                redactedEventIDs.add(it.root.redacts!!)
                redactionEventIDs.add(it.eventId)
                redactedBy[it.root.redacts!!] = it
            }
            if (it.root.isEdition()) {
                replace[it.getRelationContent()!!.eventId!!] = it
            }
        }



        fun createWrapper(event: TimelineEvent, ignoreReplace: Boolean = false): TimelineEventWrapper =
            if (redactedEventIDs.contains(event.eventId)) TimelineEventWrapper.Redacted(
                event,
                redactedBy[event.eventId]!!
            )
            else if (!ignoreReplace && replace.containsKey(event.eventId)) TimelineEventWrapper.Replaced(
                event,
                createWrapper(event, true),
                replace[event.eventId]!!
            )
            else if(event.isReply()) TimelineEventWrapper.Replied(
                event,
                event.root.getRelationContent()!!.inReplyTo!!.eventId!!
            )
            else TimelineEventWrapper.Default(event)

        var lastUserId: SenderInfo? = null
        timelineEvents.filter { !redactionEventIDs.contains(it.eventId) && !it.root.isEdition() }.forEach {
            fun isMerged(root: Event) =
                root.isTextMessage()
                        || root.isAudioMessage()
                        || root.isImageMessage()
                        || root.isVoiceMessage()
                        || root.isVideoMessage()

            fun addNewSection(event: TimelineEvent) {
                val root = event.root
                if(isMerged(root)) {
                    list.add(Pair(it.senderInfo, Either.Left(mutableListOf(createWrapper(event)))))
                } else {
                    list.add(Pair(it.senderInfo, Either.Right(createWrapper(event))))
                }
            }
            if (lastUserId != null && it.senderInfo == lastUserId) {
                val last = list.last().second
                when(last) {
                    is Either.Left -> {
                        if(isMerged(it.root)) {
                            last.value.add(createWrapper(it))
                        } else {
                            addNewSection(it)
                        }
                    }
                    is Either.Right -> addNewSection(it)
                }
            } else {
                addNewSection(it)
            }
            lastUserId = it.senderInfo
        }

        return list.reversed()
    }

    val name = room.roomSummary()?.displayName

    fun getName(lifecycleOwner: LifecycleOwner, consume: (String) -> Unit) {
        room.getRoomSummaryLive().observe(lifecycleOwner) { roomSummary ->
            val summary = roomSummary.getOrNull() ?: return@observe
            consume(summary.displayName)
        }
    }

    fun sendTextMessage(message: String) {
        Log.d("Sending Message", message)
        replyTo?.let {
            replyTo = null
            room.relationService()
                .replyToMessage(eventReplied = it, replyText = message, autoMarkdown = true)
        } ?: room.sendService()
            .sendTextMessage(text = message, autoMarkdown = true)
    }


    fun redact(event: TimelineEvent) {
        room.sendService().redactEvent(event = event.root, reason = null)
    }

    fun loadMore() {
        timeline?.paginate(Timeline.Direction.BACKWARDS, 30)
    }
}