package com.dfsek.feldspar.ui

import android.app.Activity
import android.content.Intent
import android.util.Log
import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.material.Icon
import androidx.compose.material.IconButton
import androidx.compose.material.MaterialTheme
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Email
import androidx.compose.material.icons.filled.KeyboardArrowLeft
import androidx.compose.material.icons.filled.KeyboardArrowRight
import androidx.compose.material.icons.filled.Person
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import com.dfsek.feldspar.AppState
import com.dfsek.feldspar.ui.rooms.Avatar
import com.dfsek.feldspar.ui.rooms.DirectMessagesActivity
import com.dfsek.feldspar.ui.rooms.GroupMessagesActivity
import com.dfsek.feldspar.ui.rooms.OnlineIndicator
import com.dfsek.feldspar.ui.settings.SettingsDropdown
import com.dfsek.feldspar.util.compose_fixes.FixedDefaultFlingBehavior
import com.dfsek.feldspar.util.getPreviewText
import org.matrix.android.sdk.api.query.SpaceFilter
import org.matrix.android.sdk.api.session.room.Room
import org.matrix.android.sdk.api.session.room.RoomSummaryQueryParams
import org.matrix.android.sdk.api.session.room.model.RoomSummary
import org.matrix.android.sdk.api.session.space.Space
import java.util.Stack

@Composable
fun Activity.RootSpacesSelection(modifier: Modifier = Modifier, isOpen: MutableState<Boolean>) {
    Column(modifier = modifier
        .fillMaxWidth(0.75f)
        .fillMaxHeight()
        .background(MaterialTheme.colors.background)
        .clickable(interactionSource = remember { MutableInteractionSource() }, indication = null) {}) {
        var rootSpaces by remember { mutableStateOf<List<Room>>(emptyList()) }
        AppState.session?.let { session ->
            val spaceStack = remember { Stack<Space>() }
            var currentRooms by remember { mutableStateOf<List<Room>?>(null) }
            LaunchedEffect(AppState.session) {
                rootSpaces = session
                    .spaceService()
                    .getRootSpaceSummaries()
                    .map { session.spaceService().getSpace(it.roomId)?.asRoom() as Room }
                if (currentRooms == null) currentRooms = rootSpaces
                Log.d("Root Spaces", "Populating root spaces, Size: ${rootSpaces.size}")
            }

            Row {
                IconButton(onClick = {
                    if (spaceStack.empty()) {
                        isOpen.value = false
                    } else {
                        spaceStack.pop()
                        currentRooms = if (spaceStack.empty()) {
                            rootSpaces.map { it.roomSummary() as RoomSummary }
                        } else {
                            session
                                .spaceService()
                                .getSpaceSummaries(
                                    RoomSummaryQueryParams.Builder()
                                        .apply { spaceFilter = SpaceFilter.ActiveSpace(spaceStack.peek().spaceId) }.build()
                                )
                        }.map { session.roomService().getRoom(it.roomId) as Room }
                    }
                }) {
                    Icon(
                        imageVector = Icons.Default.KeyboardArrowLeft,
                        contentDescription = "Close"
                    )
                }
            }
            val avatarSize = AppState.Preferences.spacesAvatarSize
            val fling = FixedDefaultFlingBehavior.fixedFlingBehavior()
            LazyColumn(flingBehavior = fling) {
                if(spaceStack.empty()) {
                    item {
                        Row(modifier = Modifier.fillMaxWidth().clickable {
                            runOnUiThread {
                                startActivity(Intent(applicationContext, DirectMessagesActivity::class.java))
                                finish()
                            }
                        }) {
                            Icon(
                                imageVector = Icons.Default.Person,
                                contentDescription = "Direct Messages",
                                modifier = Modifier.size(avatarSize.dp)
                            )
                            Text("Direct Messages", fontSize = 18.sp)
                        }
                    }
                    item {
                        Row(modifier = Modifier.fillMaxWidth().clickable {
                            runOnUiThread {
                                startActivity(Intent(applicationContext, GroupMessagesActivity::class.java))
                                finish()
                            }
                        }) {
                            Icon(
                                imageVector = Icons.Default.Email,
                                contentDescription = "Group Messages",
                                modifier = Modifier.size(avatarSize.dp)
                            )
                            Text("Group Messages", fontSize = 18.sp)
                        }
                    }
                }

                items(currentRooms ?: emptyList(), key = {
                    it.roomId
                }) { room ->
                    Row(modifier = Modifier.clickable {
                        room.asSpace()?.let {
                            spaceStack.push(it)
                            currentRooms = session
                                .spaceService()
                                .getSpaceSummaries(
                                    RoomSummaryQueryParams.Builder()
                                        .apply { spaceFilter = SpaceFilter.ActiveSpace(it.spaceId) }.build()
                                )
                                .filter { !it.isDirect } // don't include DMs in space previews
                                .map { session.roomService().getRoom(it.roomId) as Room }
                        } ?: startActivity(Intent(applicationContext, RoomActivity::class.java).apply {
                            putExtra("room", room.roomId)
                        })
                    }.fillMaxWidth()) {
                        val name = remember { room.roomSummary()?.displayName ?: "" }
                        val lastContent = remember { room.roomSummary()?.latestPreviewableEvent }

                        Avatar(room.roomSummary()?.avatarUrl, avatarSize, name = room.roomSummary()?.displayName)

                        Column {
                            Text(name, fontSize = 18.sp)
                            Text(lastContent?.getPreviewText() ?: "", fontSize = 12.sp)
                        }
                    }
                }
            }
        }
    }
}

@Composable
fun BackButton(modifier: Modifier = Modifier, isSelectionOpen: MutableState<Boolean>) {
    IconButton(onClick = {
        isSelectionOpen.value = true
    }, modifier = modifier) {
        Icon(
            imageVector = Icons.Default.KeyboardArrowRight,
            contentDescription = "Spaces"
        )
    }
}

@Composable
fun Activity.TopBar(
    modifier: Modifier,
    name: String = "",
    isSelectionOpen: MutableState<Boolean>,
    center: @Composable BoxScope.() -> Unit = {Text(name, fontSize = 24.sp, modifier = Modifier.align(Alignment.Center), color = MaterialTheme.colors.onSurface)}
) {
    Box (modifier = modifier
        .background(MaterialTheme.colors.surface)
        .fillMaxWidth()) {
        BackButton(modifier = Modifier.align(Alignment.CenterStart), isSelectionOpen)
        center.invoke(this)
        SettingsDropdown(modifier = Modifier.align(Alignment.CenterEnd))
    }
}

@Composable
fun Activity.SelectionUI(isSelectionOpen: MutableState<Boolean>) {
    if(isSelectionOpen.value) {
        Box(modifier = Modifier.fillMaxSize().clickable(interactionSource = remember { MutableInteractionSource() }, indication = null) {
            isSelectionOpen.value = false
        }) {
            RootSpacesSelection(modifier = Modifier, isSelectionOpen)
        }
    }
}