package com.dfsek.feldspar.util

import android.media.MediaPlayer
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.material.Icon
import androidx.compose.material.IconButton
import androidx.compose.material.MaterialTheme
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Pause
import androidx.compose.material.icons.filled.PlayArrow
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.SolidColor
import androidx.compose.ui.layout.ContentScale
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.unit.dp
import coil.compose.AsyncImage
import coil.decode.BitmapFactoryDecoder
import coil.request.ImageRequest
import com.dfsek.feldspar.AppState
import com.halilibo.richtext.markdown.Markdown
import com.halilibo.richtext.ui.RichText
import com.halilibo.richtext.ui.RichTextThemeIntegration
import com.linc.audiowaveform.AudioWaveform
import kotlinx.coroutines.delay
import linc.com.amplituda.callback.AmplitudaErrorListener
import org.matrix.android.sdk.api.session.events.model.Content
import org.matrix.android.sdk.api.session.events.model.EventType
import org.matrix.android.sdk.api.session.events.model.toModel
import org.matrix.android.sdk.api.session.room.model.*
import org.matrix.android.sdk.api.session.room.model.message.*
import org.matrix.android.sdk.api.session.room.timeline.TimelineEvent
import org.matrix.android.sdk.api.session.room.timeline.getLastEditNewContent
import org.matrix.android.sdk.api.util.ContentUtils
import java.io.File
import kotlin.time.Duration.Companion.seconds

@Composable
fun TimelineEvent.RenderMessage(modifier: Modifier = Modifier): Unit = when (root.getClearType()) {
    EventType.MESSAGE -> RenderMessageEvent(modifier = modifier)
    EventType.ENCRYPTED -> Text("Encrypted message (haven't received keys!)", color = MaterialTheme.colors.error)
    EventType.STATE_ROOM_MEMBER -> RenderMemberEvent()
    EventType.STATE_ROOM_CREATE -> Text("${senderInfo.disambiguatedDisplayName} created the room.")
    EventType.STATE_ROOM_NAME -> Text(
        "${senderInfo.disambiguatedDisplayName} changed the room name to ${
            root.getClearContent().toModel<RoomNameContent>()?.name
        }"
    )

    EventType.STATE_ROOM_ENCRYPTION -> Text("${senderInfo.disambiguatedDisplayName} enabled encryption.")
    EventType.STATE_ROOM_HISTORY_VISIBILITY -> Text(
        "${senderInfo.disambiguatedDisplayName} set message history visible to ${
            root.getClearContent().toModel<RoomHistoryVisibilityContent>()?.historyVisibility?.humanReadable()
        }."
    )

    EventType.STATE_ROOM_GUEST_ACCESS -> Text(
        "${senderInfo.disambiguatedDisplayName} ${
            root.getClearContent().toModel<RoomGuestAccessContent>()?.guestAccess?.humanReadable()
        } guest access."
    )

    EventType.STATE_ROOM_JOIN_RULES -> Text(
        "${senderInfo.disambiguatedDisplayName} set join rule to ${
            root.getClearContent().toModel<RoomJoinRulesContent>()?.joinRules?.humanReadable()
        }."
    )

    else -> Text("Feldspar unimplemented event ${root.getClearType()}", color = MaterialTheme.colors.error)
}

fun RoomJoinRules.humanReadable() = when (this) {
    RoomJoinRules.PUBLIC -> "public"
    RoomJoinRules.INVITE -> "invite only"
    RoomJoinRules.KNOCK -> "knock"
    RoomJoinRules.PRIVATE -> "private"
    RoomJoinRules.RESTRICTED -> "restricted"
}

fun GuestAccess.humanReadable() = when (this) {
    GuestAccess.Forbidden -> "disabled"
    GuestAccess.CanJoin -> "allowed"
}

fun RoomHistoryVisibility.humanReadable() = when (this) {
    RoomHistoryVisibility.WORLD_READABLE -> "everyone"
    RoomHistoryVisibility.SHARED -> "shared"
    RoomHistoryVisibility.INVITED -> "invited"
    RoomHistoryVisibility.JOINED -> "joined"
}

fun Membership.humanString() =
    when (this) {
        Membership.NONE -> null
        Membership.INVITE -> "was invited"
        Membership.JOIN -> "joined"
        Membership.KNOCK -> "knocked"
        Membership.LEAVE -> "left"
        Membership.BAN -> "was banned"
    }

@Composable
fun TimelineEvent.RenderMemberEvent() {
    val content = root.getClearContent().toModel<RoomMemberContent>()
    val prevContent = root.resolvedPrevContent().toModel<RoomMemberContent>()

    val isMembershipChanged = content?.membership != prevContent?.membership

    val isJoin = isMembershipChanged && content?.membership == Membership.JOIN
    val isLeave = isMembershipChanged && content?.membership == Membership.LEAVE && root.stateKey == root.senderId

    val isProfileChanged = !isMembershipChanged && content?.membership == Membership.JOIN
    val isDisplaynameChange = isProfileChanged && content?.displayName != prevContent?.displayName
    val isAvatarChange = isProfileChanged && content?.avatarUrl !== prevContent?.avatarUrl

    val text = if (isJoin) {
        "joined"
    } else if (isLeave) {
        "left"
    } else if (isDisplaynameChange) {
        "changed their name to ${content?.displayName}"
    } else if (isAvatarChange) {
        "changed their avatar"
    } else {
        content?.membership?.humanString()
    }

    val username = content?.displayName
    Text("$username $text.")
}

@Composable
fun Content.RenderContent(modifier: Modifier = Modifier) {
    val messageContent = toModel<MessageContent>() ?: return
    when (messageContent.msgType) {
        MessageType.MSGTYPE_TEXT -> RichTextThemeIntegration(contentColor = { MaterialTheme.colors.onBackground }) {
            RichText(modifier = modifier) {
                Markdown(content = formatMessage(this@RenderContent))
            }
        }

        MessageType.MSGTYPE_AUDIO -> {
            val audioContent = toModel<MessageAudioContent>() ?: return

            var audio by remember { mutableStateOf<File?>(null) }

            LaunchedEffect(audioContent) {
                try {
                    audio = AppState.session!!.fileService().downloadFile(audioContent)
                } catch (e: IllegalArgumentException) {
                    e.printStackTrace()
                }
            }

            audio?.let {
                Row(modifier = Modifier.fillMaxWidth()) {
                    val player = remember {
                        val p = MediaPlayer()
                        p.setDataSource(it.absolutePath)
                        p.prepare()
                        p
                    }
                    var waveformProgress by remember { mutableStateOf(0F) }
                    LaunchedEffect(Unit) {
                        while (true) {
                            waveformProgress = player.currentPosition.toFloat().div(player.duration)
                            delay(1.seconds / 30)
                        }
                    }


                    IconButton(onClick = {
                        if (player.isPlaying) {
                            player.pause()
                        } else {
                            player.start()
                        }
                    }) {
                        if (player.isPlaying) {
                            Icon(
                                imageVector = Icons.Default.Pause,
                                contentDescription = "Pause"
                            )
                        } else {
                            Icon(
                                imageVector = Icons.Default.PlayArrow,
                                contentDescription = "Play"
                            )
                        }
                    }

                    AudioWaveform(
                        amplitudes = AppState.amplituda!!.processAudio(it)
                            .get(AmplitudaErrorListener { it.printStackTrace() }).amplitudesAsList(),
                        progress = waveformProgress,
                        waveformBrush = SolidColor(MaterialTheme.colors.secondary),
                        progressBrush = SolidColor(MaterialTheme.colors.primary),
                        spikeWidth = 2.dp,
                        onProgressChange = {
                            player.seekTo(player.duration.times(it).toInt())
                            player.pause()
                            waveformProgress = it
                        }
                    )
                }
            } ?: Text("Rendering audio player...")
        }

        MessageType.MSGTYPE_IMAGE -> {
            val imageContent = toModel<MessageImageContent>() ?: return

            var image by remember { mutableStateOf<File?>(null) }

            LaunchedEffect(imageContent) {
                try {
                    image = AppState.session!!.fileService().downloadFile(imageContent)
                } catch (e: IllegalArgumentException) {
                    e.printStackTrace()
                }
            }

            image?.let {
                AsyncImage(
                    model = ImageRequest.Builder(LocalContext.current)
                        .data(it)
                        .crossfade(true)
                        .decoderFactory(BitmapFactoryDecoder.Factory())
                        .build(),
                    contentScale = ContentScale.Fit,
                    contentDescription = null,
                    modifier = modifier.fillMaxWidth().padding(PaddingValues(end = 24.dp))
                )
            } ?: Text("Rendering image...")
        }

        else -> Text(formatMessage(this), modifier = modifier)
    }
}

@Composable
private fun TimelineEvent.RenderMessageEvent(modifier: Modifier = Modifier) {
    getLatestContent()?.RenderContent(modifier)
}

fun TimelineEvent.getPreviewText(cut: Boolean = true): String {
    val fullText = when (root.getClearType()) {
        EventType.MESSAGE -> getLatestContent()!!.toModel<MessageContent>()?.body ?: ""
        EventType.ENCRYPTED -> "Encrypted message (haven't received keys!)"

        else -> "Feldspar unimplemented event ${root.getClearType()}"
    }
    if (!cut) return fullText

    val beforeNewline = fullText.substringBefore("\n")

    return if (beforeNewline.length <= 50) beforeNewline else (beforeNewline.substring(0, 51).trim() + "...")
}

private fun formatMessage(timelineEvent: Content): String {
    val messageContent = timelineEvent.toModel<MessageTextContent>()
        ?: return timelineEvent.toModel<MessageContent>()?.body
            ?: return ""
    return messageContent.removeInReplyFallbacks()
}

fun MessageTextContent.removeInReplyFallbacks() = ContentUtils.extractUsefulTextFromReply(body)

fun TimelineEvent.getLatestContent(): Content? = getLastEditNewContent() ?: root.getClearContent()

fun TimelineEvent.getLatestTextCleaned() = getLatestContent()?.toModel<MessageTextContent>()?.removeInReplyFallbacks()